# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
The cadcutils package contains libraries commonly use in CADC applications.

The main modules and packages in cadcutils are:

    -utils: collection of utilities for time formats, application logging
            and command line argument parsing
    -net: network utilities: authentication, Web service base client, etc
    -exceptions: exceptions thrown by the cadcutils libraries

"""

try:
    _PACKAGE_SETUP_
except NameError:
    # do any package imports here
    pass

